(* $Id: domain.ml,v 1.4.2.10 2013/10/02 15:29:40 dmasse Exp $ *)

(** Libpoly-based abstract domains *)

open Syntax
(* open Libpoly *)

(** numerical domain for quoted and unquoted variables 
    operations :
	convert_state : (float array * float * Syntax.typ_cv) array ->
                          translate an array of unquoted constraints
       				return the polyhedra (Pol.t ?)
	convert_init : (float array * float * Syntax.typ_cv) array ->
			  convert the initial constraints to constraints
				return the constraints (Cstr.t list ?)
	convert_trans : (float array * float * Syntax.typ_cv) array ->
			  convert the loop constraints to constraints
				return the constraints quoted and unquoted
						(Cstr.t list ?)
	apply_init : polyhedra -> constraints -> polyhedra
	apply_trans : polyhedra -> constraints -> polyhedra
	apply_trans_bwd : polyhedra -> constraints -> polyhedra
	join_list : polyhedra list -> polyhedra
	leq_state : polyhedra -> polyhedra -> bool
*)
exception PolEmpty
exception NotForall
class domain nbvars split varnames =
	(* nbvars : number of variables *) (* varnames : int -> string name *)
	let get_name_var vr =
		varnames ((Var.toInt vr)-1)
	in
	(* construct the liste of unquoted and quoted vars *)
	let create_list n f =
		    let rec create_list_p p l =
			if p>=n then l else
			     create_list_p (p+1) ((f p)::l)
		    in create_list_p 0 []
	in 
	let unqtdvars = create_list nbvars (fun i -> Var.fromInt (i+1))
	and qtdvars = create_list nbvars (fun i -> Var.fromInt (i+nbvars+1))
	and assign_qtd_to_unqtd =
		create_list nbvars (fun i -> 
			{ Pol.var = Var.fromInt (i+1) ;
			  Pol.lin = [ Num.u , Var.fromInt (i+nbvars+1) ];
			  Pol.cst = Num.z })
	and assign_unqtd_to_qtd =
		create_list nbvars (fun i -> 
			{ Pol.var = Var.fromInt (i+nbvars+1) ;
			  Pol.lin = [ Num.u , Var.fromInt (i+1) ];
			  Pol.cst = Num.z })
	in
	let convert_cstr_of_intern (a,ct,tp) =
	    let pcl a =
		let lst = ref [] in
		Array.iteri (fun i v ->
				if v<>0.0 then
			lst := (Num.of_float v, Var.fromInt (i+1))::!lst) a;
		!lst 
	    in Cstr.mk (match tp with
			  EQ -> Cstr.Eq
			| LESSEQ -> Cstr.Le
			| LESS -> Cstr.Lt) (pcl a) (Num.of_float (-.ct))
	in
(* 	let convert_cstr_to_intern ct =
	    let linterm = ct.Cstr.v in
	    let coefs = List.map (fun v -> Vec.get linterm v) unqtdvars
	    in (Array.of_list coefs,(Num.neg ct.Cstr.c),
			(match ct.Cstr.typ with
			   Cmp.Eq -> EQ
			 | Cmp.Le -> LESSEQ
			 | Cmp.Lt -> failwith "strict constraint"))
	in *)
	let rec inter_list_consts pol_opt lconsts =
	    match lconsts with 
	       [] -> pol_opt
	     | cst1::tlcst -> begin
			        match pol_opt with
				  None -> None
				| Some p -> let res,_ = Pol.add p cst1 in
						inter_list_consts res tlcst
			      end
	in
	let lsplit = List.length split in
	let nb_domains_split = 1 lsl lsplit in
	let list_split_cstrs = List.map convert_cstr_of_intern split in
        let array_splits =
	    Array.init nb_domains_split
		(fun i -> List.mapi (fun j cst ->
				if ((i lsr j) land 1 = 0) then
					cst else Cstr.compl cst) 
						list_split_cstrs) 
in
object (self:'self)
	val nbvars = nbvars
	val array_splits = array_splits
	val nb_splits = nb_domains_split

	method get_array_splits = array_splits
	
	method convert_init initconsts =
		List.map convert_cstr_of_intern initconsts

	method convert_trans transconsts =
		List.map convert_cstr_of_intern transconsts
	
	method split_one_poly pol1 = 
		match pol1 with
                  None -> Array.make nb_splits None
                | Some pl -> Array.map (fun lc ->
                                   inter_list_consts pol1 lc) array_splits
				(* FIXME : not efficient n.2^n constraints *)

	method list_consts_to_state listconsts =
		let pol1 = Pol.mk listconsts in
		self#split_one_poly pol1

	method top =
		Array.map (fun lc -> Pol.mk lc) array_splits

	method copy_state st = Array.copy st

	method convert_state initstate =
	    	self#list_consts_to_state (self#convert_init initstate)

	method apply_init state consts =
		Array.map (fun st -> inter_list_consts st consts) state

	method join_state state1 state2 =
		Array.mapi (fun i pol1 ->
			match pol1,state2.(i) with
			  None,_ -> state2.(i)
			| _,None -> pol1
			| Some p1, Some p2 -> let rp,_ = Pol.join p1 p2 in
						Some rp) state1

	method join_state_i state1 state2 =
		for i=0 to nb_splits-1 do
		    state1.(i) <- 
			begin
			match state1.(i),state2.(i) with
			  None,a -> a
			| a,None -> a
			| Some p1, Some p2 ->  Some (fst (Pol.join p1 p2))
			end
		done

	method apply_trans_one_poly poly consts =
		(* fist, intersect with the consts *)
		(* then, assign the quoted variables to the unquoted ones *)
		(* finally, forget the quoted variables *)
		(* TODO : use InEqs and Eqs.rename ??? *)
		let rt = inter_list_consts poly consts
		in 
		let res = match rt with
		     None -> None
		   | Some p ->
			let asp = Pol.assignM p assign_qtd_to_unqtd in
			    Some (fst (Pol.projectM asp qtdvars))
		in self#split_one_poly res

	method apply_trans state consts =
		let res1 = self#apply_trans_one_poly state.(0) consts in
		for j=1 to nb_splits-1 do
		    let res = self#apply_trans_one_poly state.(j) consts in
		    self#join_state_i res1 res
		done;
		res1

	method apply_trans_one_poly_bwd poly consts =
		(* fist, assign the unquoted variables to the quoted ones *)
		(* then, forget the unquoted variables *)
		(* then intersect with the consts *)
		(* finally, forget the quoted variables *)
		(* TODO : use InEqs and Eqs.rename ??? *)
		let res =
		match poly with
		  None -> None
		| Some p -> let asp = Pol.assignM p assign_unqtd_to_qtd in
			    let asp2 = Some (fst (Pol.projectM asp unqtdvars))
			    in begin
			       let rt = inter_list_consts asp2 consts
			       in match rt with
				    None -> None
				  | Some p2 -> Some 
					(fst (Pol.projectM p2 qtdvars))
			       end
		in self#split_one_poly res

        method apply_trans_bwd state consts =
		let res1 = self#apply_trans_one_poly_bwd state.(0) consts in
                for j=1 to nb_splits-1 do
                    let res = self#apply_trans_one_poly_bwd state.(j) consts in
                    self#join_state_i res1 res
                done;
                res1

	method join_list state_list =
		let res1 = self#copy_state (List.hd state_list) in
		List.iter (fun res -> self#join_state_i res1 res) 
					(List.tl state_list);
		res1

	method leq_poly st1 st2 =
(*		Configuration.debugf "leq poly %a %a@." self#print_poly st1 self#print_poly st2; *)
		match st1,st2 with
		  None,_ -> true
		| _,None -> false
		| Some pol1,Some pol2 ->
			(Pol.isIncl pol1 pol2 <> Pol.NoIncl)

	method leq_state st1 st2 =
		try
			Array.iteri (fun i p1 -> 
				if (not (self#leq_poly p1 st2.(i))) then
					raise NotForall) st1;
			true
		with
			NotForall -> false

	method widen_poly st1 st2 =
                match st1,st2 with
                  None,_ -> st2
                | _,None -> st1
                | Some pol1,Some pol2 -> Some (Pol.widen pol1 pol2)

	method widen_state st1 st2 =
		Array.mapi (fun i p -> 
				let rp = self#widen_poly p st2.(i)
				in inter_list_consts rp array_splits.(i)) st1

	(* translate non-empty polyhedra to constraints *)
	method poly_to_linconsts pol1 =
		(let eq1 = pol1.Pol.eqs in 
				List.map (fun (_,x) -> x.Cons.c) eq1)@
			       (let eq2 = pol1.Pol.ineqs in
				List.map (fun x -> x.Cons.c) eq2)

	method affect_linconsts_to_split st =
		Array.iteri (fun i pl ->
				array_splits.(i) <-
				match pl with
				  None -> 
					[ Cstr.mk Cstr.Le [] Num.negU ]
				| Some p -> self#poly_to_linconsts p) st

	(* translate states to template *)
	method poly_to_template st1 = 
		match st1 with
		  None -> None,[]
		| Some pol1 ->
			let lconsts = self#poly_to_linconsts pol1 in
			let rec pl_split l = 
				match l with
				     [] -> [],[]
				   | a::tll -> let tp1,pf1 = pl_split tll in
					begin
					  if Cstr.get_typ a = Cstr.Eq then
					  begin
					     let b1,b2 = Cstr.split a in
					     b1.Cstr.v::(b2.Cstr.v::tp1),
					     b1.Cstr.c::(b2.Cstr.c::pf1)
					  end
					  else 
					     a.Cstr.v::tp1 , a.Cstr.c::pf1
					end
			in let (u,v)= pl_split lconsts in (Some u,v)

	method state_to_template st =
		let lgt = Array.length st in
		let tb = Array.make lgt None
		and tfp = Array.make lgt [] in
		Array.iteri (fun i a -> 
			let r1,r2 = self#poly_to_template a
			in tb.(i) <- r1; tfp.(i) <- r2) st;
		tb,tfp

	method template_to_poly template pfp splt =
		try
		   match Pol.mk splt with
		     None -> raise PolEmpty
		   | Some deb -> begin
		       Some (List.fold_left2 (fun pol vec cst ->
		 	    if (Num.cmp cst (Num.of_float (-.infinity))=0) 
						then raise PolEmpty; 
					(* FIXME *)
			    let cstr = { Cstr.typ = Cstr.Le; Cstr.v = vec;
                                        Cstr.c = cst } in
			    match fst (Pol.add pol cstr) with
			      None -> raise PolEmpty
			    | Some a -> a) deb template pfp)
		     end
		with
		   PolEmpty -> None

        method template_to_state template_base sol =
		match sol with
		  None -> Array.map (fun _ -> None) template_base
		| Some esol->
			Array.mapi (fun i u ->
				match u with
				  None -> None
				| Some a -> self#template_to_poly 
					      a 
					      (Array.to_list esol.(i)) 
					      array_splits.(i)) template_base

	method print_poly fmt st =
		match st with
		  None -> Format.pp_print_string fmt "bot"
		| Some pol -> Format.pp_print_string fmt 
					(Pol.pr get_name_var pol) 
		
	method print_state fmt st =
		Format.pp_print_string fmt "[ ";
		Array.iter (fun a -> Format.fprintf fmt "%a ;" 
					self#print_poly a) st;
		Format.fprintf fmt "]@."

end
