(** forward policy iteration to compute lfp, simple version *)

(*
open Mlglpk
open Syntax

(* structures *)

(** rhs of an equation for ONE transition (no max) *)
(* with strategy selection *)
type rhs_trans = 
	   Rhs_lp of float ref * Mlglpk.mlglp_prob * float array * 
					float array 
			(* optimal sol, lp, template line, partial rhs *)
			(* mlglp_prob : max fx | ( A ) (x / x') <= (q / \rho) *)
	 | Rhs_cst of float				(* constant *)

let print_rhs fmt l =
	match l with
	  Rhs_cst x -> Format.fprintf fmt "@ cst(%f)@ " x
	| Rhs_lp (_,lp,_,_) -> Format.fprintf fmt "@ (%a)@ " 
				Mlglpk_aapi.print_problem lp
	
(** utilities for matrices *)
let fill_matrix mat f bi bj ni nj =
	for i = 0 to ni-1 do
	for j = 0 to nj-1 do
		mat.(bi+i).(bj+j) <- f i j
	done;
	done
(* let transpose_matrix mat ni nj =
	Array.init ni
			(fun i -> Array.init nj (fun j -> mat.(j).(i)))
*)

let rec list_map_filter f l = match l with
				[] -> []
			      | a::b -> begin
					  match f a with
					    None -> list_map_filter f b
					  | Some u -> u::(list_map_filter f b)
					end

(** equation construction, from nb variables, program loop, postfp element, and
    template : 
    the result is a list (for each transition) of array (for each variable) template *)
(* loop : (float array * float * type_cv) array list *)
(* prefp : float array  (assuming we already know the values in the
   template *)
(* template : float array array *)
let construct_equation_set nbvar loop prefp template = 
	(* construct a list, with one array for each transition
	   separately (some mlglp_prob are identical) *)
	(* the first element of the list gives a global minimum *)
	(* hyp : no element of postfp is -oo *)
	Configuration.debugf "[fwdPI] construct_equation_set %d@." nbvar;
	let i = ref 0 in
	(Array.map (fun x -> Rhs_cst x) prefp)::
	list_map_filter 
		(fun trans -> 
			(* each transition is an array of contraints *)
			(* construction of (   Q    )
					   ( T   0  ) *)
			let strans = Array.length trans and
			    stemplate = Array.length template in
			let cmat = 2*nbvar 
			and lmat = strans + stemplate in
			let mat = Array.make_matrix lmat cmat 0.0 in
			let rhs_part =
				Array.map (fun (_,b,_) -> (-.b))
				 trans
			in
			fill_matrix mat 
				(fun i j -> let (a,_,_) = trans.(i) 
						in a.(j)) 0 0 strans (nbvar*2);
			fill_matrix mat
				(fun i j -> template.(i).(j)) 
				strans 0 stemplate nbvar;
			let solpb = (* not empty problem *)
			Mlglpk_aapi.construct_problem 
			  ~name:("LP_"^(string_of_int !i))
			  ~direction:GLP_MAX
			  ~matrix:mat lmat cmat in
			Array.iteri (fun j (_,b,t) ->
				Mlglpk.glp_set_row_bnds solpb 
					(j+1) 
					(if t = EQ then GLP_FX (-.b)
						   else GLP_LO (-.b))) trans;
			Array.iteri (fun j x ->
				if x < infinity then
				    Mlglpk.glp_set_row_bnds solpb
					 (strans+j+1) (GLP_UP x)
				    else Mlglpk.glp_set_row_bnds solpb
                                         (strans+j+1) GLP_FR) prefp;
			for j = 0 to nbvar*2 -1 do
				Mlglpk.glp_set_col_bnds solpb
					(j+1) GLP_FR
			done;
			(*  *)
			glp_scale_prob solpb [GLP_SF_AUTO];
			glp_adv_basis solpb;
			glp_simplex solpb None; (* TODO : check exception? *)
			(*  *)
			let compute_sol = (glp_get_status solpb = GLP_OPT) in
			if (not compute_sol) && 
				Configuration.fwd_PI_lookahead#get 
			then None else
			Some begin
			  Array.init stemplate 
				(fun k ->
					let tline = Array.init nbvar 
					  (fun j -> 
					   let x = template.(k).(j) in
					   if compute_sol then
					       Mlglpk.glp_set_obj_coef solpb
						    (nbvar+j+1) x;
					   x) in
					let optval = 
					begin
					   if compute_sol then
					   begin
					     (*   *)
					     Mlglpk.glp_simplex solpb None;
					     (*   *)
					     if glp_get_status solpb = GLP_OPT
					     then
						glp_get_obj_val solpb
					     else infinity
					   end else -.infinity
					end in
	   				Rhs_lp 
					    (ref optval,
					     solpb, tline, rhs_part)
				)
			end 
		) 
		loop

(** improving the strategy, recomputing the linear programs :
    take the set of equations, the local results, and modify the strategy
    represented as an array (for each local variable) of transitions *)
let improve_strat nbvar eqs sol strat =
	Configuration.debugf "[fwdPI] improve_strat@.";
	let nbv = Array.length sol in
	let global_changed = ref false in
	List.iter (
		fun tab -> let changed_lp = ref false in
		Array.iteri
		(fun i u -> 
		  let curval = sol.(i) in
		  match u with
		    Rhs_cst x -> if x>curval then
				 begin
				   global_changed := true;
				   sol.(i) <- x;
				   strat.(i) <- u
				 end
		  | Rhs_lp (fr,lp,tmpl,rp) ->
		      begin
			let nbrlp = Mlglpk.glp_get_num_rows lp in
			if (not !changed_lp) then
			begin
			  Array.iteri (fun j x ->
			  	if x < infinity then
				    Mlglpk.glp_set_row_bnds lp
					 (nbrlp-nbv+j+1) (GLP_UP x)
				    else Mlglpk.glp_set_row_bnds lp
                                         (nbrlp-nbv+j+1) GLP_FR) sol
			end;
			changed_lp := true;
			Array.iteri (fun j x ->
				    Mlglpk.glp_set_obj_coef lp 
						(nbvar+j+1) x) tmpl;
			Mlglpk.glp_simplex lp None;
			let nv = 
			begin
			     match glp_get_status lp with
			       GLP_OPT -> glp_get_obj_val lp
			     | GLP_UNBND -> infinity
			     | GLP_NOFEAS -> -.infinity
			     | _ -> failwith "not usable status"
			end in
			if nv > curval +. Configuration.epsilon then 
						(* TODO : check *)
			begin
			   fr := nv;
			   global_changed := true;
			   sol.(i) <- nv;
			   strat.(i) <- u
			end
		      end
		   ) tab
		) eqs;
	 !global_changed


(* 
let find_actual_value r =
	match r with
	  Rhs_cst x -> x
	| Rhs_lp (a,_,_,_,_) -> !a

let list_map_option f l = 
	let rec parc lst =
	   match lst with
	     [] -> []
	   | a::b -> begin
		   match f a with
		     None -> parc b
		   | Some y -> y::(parc b)
		     end
	in parc l
*)

(* construct local strategy equation system using the global system *)
let construct_local_strat nbvar strat =
	Configuration.debugf "[fwdPI] construct local strat@.";
	let nbv = Array.length strat in
	let linpg = 
		Mlglpk_aapi.construct_problem
			~direction:GLP_MAX
			~colbnds:(Array.make nbv GLP_FR)
			~objvect:(Array.make nbv 1.0)
			0 nbv in
        Array.iteri (
		fun actvar cs -> begin
			match cs with
		          Rhs_cst x -> 
				begin
				  let numr = Mlglpk.glp_add_rows linpg 1 in
				  Mlglpk.glp_set_row_bnds linpg numr (GLP_UP x);
				  Mlglpk.glp_set_mat_row linpg numr
						 [|actvar+1|] [|1.0|]
				end
			| Rhs_lp (_,lp,tmpl,_) ->
				begin
				  let begx = Mlglpk.glp_add_cols linpg 
							 (2*nbvar) in
				  for i =0 to 2*nbvar-1 do
					Mlglpk.glp_set_col_bnds linpg
						(begx+i) GLP_FR
				  done;
				  let numconsts = Mlglpk.glp_get_num_rows
						lp in
				  let begconsts = Mlglpk.glp_add_rows linpg
								numconsts in
				  for i=0 to numconsts-1 do
					let id,vl =
						Mlglpk.glp_get_mat_row lp
						(i+1) in
					let mid = Array.map
                                                  (fun i -> i+begx-1) id in
					let b = Mlglpk.glp_get_row_bnds
                                                        lp (i+1) in
					if (i<numconsts-nbv) ||
						(b= GLP_FR) then
					begin
					  Mlglpk.glp_set_row_bnds linpg
						(begconsts+i) b;
					  Mlglpk.glp_set_mat_row linpg
						(begconsts+i) mid vl
					end else
					begin
					  match b with
					   GLP_UP _ ->
					    begin
						Mlglpk.glp_set_row_bnds linpg
                                                (begconsts+i) (GLP_UP 0.0);
						Mlglpk.glp_set_mat_row linpg
							(begconsts+i)
				   (Array.append mid [|i-(numconsts-nbv)+1|])
				   (Array.append vl [|-1.0|])
					    end
					  | _ ->
						failwith "[fwdPI] construct_local_strat : impossible bound"
					end
				  done;
				  let newconst = Mlglpk.glp_add_rows linpg 1 in
				  Mlglpk.glp_set_row_bnds linpg newconst
						 (GLP_LO 0.0);
				  Mlglpk.glp_set_mat_row linpg newconst
					(Array.init (1+nbvar) 
					    (fun i -> if i=nbvar then (actvar+1)
						else (begx+nbvar+i)))
					(Array.init (1+nbvar) 
					    (fun i -> if i=nbvar then (-1.0)
						else tmpl.(i)))
				end
			end
		) strat;
		(* construct projective linear program *)
		let plinpg = Mlglpk.glp_create_prob () in
		begin
		   Mlglpk.glp_copy_prob 
				plinpg linpg false;
		   (* add col bnds *)
		   for i=1 to (Mlglpk.glp_get_num_cols linpg) do
			Mlglpk.glp_set_col_bnds plinpg
				i (GLP_DB (0.0,1.0))
		   done;
		   (* delete rhs *)
		   for i=1 to (Mlglpk.glp_get_num_rows linpg) do
			Mlglpk.glp_set_row_bnds plinpg
                                i 
				( match Mlglpk.glp_get_row_bnds linpg i with
					GLP_FR -> GLP_FR
				      | GLP_LO _ -> GLP_LO 0.0
				      | GLP_UP _ -> GLP_UP 0.0
				      | GLP_DB (_,_) -> GLP_FX 0.0
				      | GLP_FX _ -> GLP_FX 0.0	
				)
                   done
		end;
		Configuration.debugf "[fwdPI] local strat constructed : %a@,%a@." Mlglpk_aapi.print_problem linpg Mlglpk_aapi.print_problem plinpg;
		linpg,plinpg
	

exception SolInfinity
(** solving local strategy (really bad approach) *)
let solve_local_strat (strat,pstrat) res =
	let nbv = Array.length res in
	Configuration.debugf "[fwdPI] solve_local_strat, pstrat first, nbv=%d@." nbv;
 	(* *)
        Mlglpk.glp_scale_prob pstrat [GLP_SF_AUTO];
        Mlglpk.glp_adv_basis pstrat;
	Mlglpk.glp_simplex pstrat None;
	(* *)
	for i = 1 to Mlglpk.glp_get_num_cols pstrat do
	   if Mlglpk.glp_get_col_prim pstrat i > 0.0 then
	   begin
	      Configuration.debugf ~level:6 "[fwdPI,solve] found infinite col : %d@." i;
	      Mlglpk.glp_set_col_bnds strat i (GLP_FX 0.0);
	      if (i<=nbv) then res.(i-1) <- infinity
	   end
	done;
 	(* *)
        Mlglpk.glp_scale_prob strat [GLP_SF_AUTO];
        Mlglpk.glp_adv_basis strat;
	Mlglpk.glp_simplex strat None;
	(* *)
	let notinf = ref false in
	for i = 1 to nbv do
	   if res.(i-1) <> infinity then
	   begin
		res.(i-1) <- Mlglpk.glp_get_col_prim strat i;
		notinf := true
	   end
	done;
	!notinf
	(* this is a very bad work *)


(* general policy iteration *)
let lfp_fwdPI nbv loop prefp template =
	let nb_it = ref 1 in
	try
	  let eqs = construct_equation_set nbv loop prefp template in
	  let sol = Array.copy prefp in
          let policy = List.hd eqs in
	  let unstable = ref (improve_strat nbv eqs sol policy) in
	  while ((!unstable) && (!nb_it <= Configuration.fwd_PI_max_iterations#get)) do 
	     Configuration.debugf "[fwdlfp PI] : iteration %d (sol %a) (policy %a)@." (!nb_it) (fun fmt a -> Array.iter (fun elt -> Format.fprintf fmt "%g " elt) a) sol (fun fmt a -> Array.iter (print_rhs fmt) a) policy;
	     let strats = construct_local_strat nbv policy in
	     Configuration.debugf "[fwdlfp PI] : strat constructed@." (* (fun fmt a -> Array.iteri (fun i lst -> List.iter 
		(fun (elt,cst) -> Format.fprintf fmt "v%d >= %a+%f , " i
			(fun fmt u -> Array.iteri (fun j coef
			-> Format.fprintf fmt "%fv%d+" coef j) u) elt cst)
				 lst) a) strat *) ;
	     if not (solve_local_strat strats sol) then
		raise SolInfinity;
	     Configuration.debugf "[fwdlfp PI] : end solving strat (sol %a)@." (fun fmt a -> Array.iter (fun elt -> Format.fprintf fmt "%g " elt) a) sol;
	     unstable := improve_strat nbv eqs sol policy;
	     nb_it := !nb_it + 1
          done;
	  sol
	with
	  SolInfinity -> begin
			   Configuration.debugf "[fwdlfp PI] : infinite return@.";
			   Array.make (Array.length prefp)  infinity
			 end
*)
