(** syntax.ml : syntax of the small language *)

(** variables *)
type variable = int

let nb_variables = ref 0
let var_names = ref [||]
let get_var_name i = (!var_names).(i)

let declare_variables l_names =
		var_names := Array.of_list l_names;
		Array.fast_sort Pervasives.compare (!var_names);
		nb_variables := Array.length (!var_names);
		for i=0 to (!nb_variables - 2) do
		   if (!var_names.(i) = !var_names.(i+1)) then
		   begin
			Configuration.errorf "Two identical variables : %s@."
				(!var_names.(i));
			failwith "declare_variables: identical variables"
		   end
		done;
		Configuration.debugdo ~level:2 
			(fun () -> 
				Configuration.outf "%d variables :"
					(!nb_variables);
				Array.iter (fun n -> Configuration.outf " %s"
							n) (!var_names);
				Configuration.outf "@.")
		

exception Found of int
let dicho_find s tab =
	try 
		let lb = ref 0 and ub = ref (Array.length tab-1) in
		while (!ub >= !lb) do
		   let mid = (!ub + !lb) /2 in
		   if tab.(mid) = s then
			raise (Found mid)
		   else if tab.(mid) < s then 
			lb := mid +1
		   else ub := mid -1
		done;
		raise Not_found
	with
		Found i -> i
			

let get_id_var name = dicho_find name (!var_names)
		
(** concrete syntax *)
type typ_cv = EQ | LESSEQ | LESS
type monome = Var of float * int
	   | VarPrimed of float * int
	   | Cst of float
type expr = monome list
type concrete_const = expr * typ_cv

let double_expr x = float_of_string x
let integer_expr x = float_of_string x

let expr_of_var name primed = 
	try 
		let id = get_id_var name in
		   if primed then [ VarPrimed (1.0,id) ]
		   else [ Var (1.0,id) ]
	with 
		Not_found -> begin
			Configuration.errorf "Unkwown variable : %s@."
				name;
			failwith "expr_of_var : unknown variable"
			     end
let prod_expr flt exp = List.map
			(fun m -> match m with
				    Var (a,b) -> Var (flt*.a, b)
				  | VarPrimed (a,b) -> VarPrimed (flt*.a, b)
				  | Cst a -> Cst (flt*.a)) exp
let div_expr exp flt = if (flt = 0.0) then
				begin
				Configuration.errorf "Division by 0@.";
				failwith "div_expr : division by 0"
				end else
			prod_expr (1.0 /. flt) exp
let uminus_expr exp = prod_expr (-1.0) exp
let plus_expr exp1 exp2 = 
			exp2 @ exp1
let minus_expr exp1 exp2 = 
			(uminus_expr exp2) @ exp1
let cst_expr flt = [Cst flt]

(** constraints on variables (with primed) *)
type cv = float array * float * typ_cv 
			       (* 2* !nb_variables size *)
			       (* ou !nb_variable size pour l'initial
				  ou les contraintes courantes *)
			       (* pb utilisation float ? *)
type transition = cv list  (* taille quelconque *)

(* translating concrete constraints to abstract constraints *)
let check_init lconsts =
	Configuration.debugf 
			~level:9 "check_init : %d\n" (List.length lconsts);
	List.map (fun (lmon,tp) ->
		let tab = Array.make !nb_variables 0.0 and
		    cst = ref 0.0 in
		List.iter (fun m -> match m with
		     Var (fl,id) ->
			tab.(id) <- tab.(id) +. fl
		   | VarPrimed (fl,id) ->
			begin
			Configuration.errorf 
			  "Primed variable in initializations: %s@." 
				(get_var_name id);
			failwith "check_init : primed variable"
			end
		   | Cst fl -> cst := !cst +. fl) lmon;
		(tab,!cst,tp)) lconsts
let check_trans ?(split=false) lconsts =
	Configuration.debugf 
			~level:9 "check_trans : %d@." (List.length lconsts);
	List.map (fun (lmon,tp) ->
		if (split && (tp=EQ)) then
		begin
		  Configuration.errorf "equality split not allowed !@.";
		  failwith "check_trans: eqsplit"
		end;
		let tab = Array.make (2* !nb_variables) 0.0 and
		    cst = ref 0.0 in
		List.iter (fun m -> match m with
		     Var (fl,id) ->
			tab.(id) <- tab.(id) +. fl
		   | VarPrimed (fl,id) ->
			tab.(id+ !nb_variables) <- tab.(id+ !nb_variables) +. fl
		   | Cst fl -> cst := !cst +. fl) lmon;
		(tab,!cst,tp)) lconsts
let distribute_trans lst_t1 trans_lists =
	if trans_lists = [] then [ lst_t1 ] else
	List.map (fun at2 -> List.append lst_t1 at2) trans_lists
				

(* let construct_template_from_state ar =
	(* construct the template as a list *)
	let templst,cstlst = Array.fold_right (fun line (tlst,clst) ->
			match line with
			  (ln,cst,LESSEQ) ->
				(Array.copy ln)::tlst,
				cst::clst
			| (ln,cst,EQ) ->
				(Array.map (fun x -> -.x) ln)::
					(Array.copy ln)::tlst,
				cst::(-.cst)::clst
			) ar ([],[])
	in (Array.of_list templst),(Array.of_list cstlst)
	
let return_state_from_template temp cst =
	Array.mapi (fun i line ->
		if (cst.(i) = infinity) then 
			(Array.map (fun x -> 0.0) line,-1.0,LESSEQ)
			else if (cst.(i) = -.infinity) then
			(Array.map (fun x -> 0.0) line,1.0,LESSEQ)
			else
			(Array.copy line,-.cst.(i),LESSEQ)) temp *)

type prog = transition * transition * transition list
		(* 1ère transition : initialisation *)
		(* liste : boucle *)

let get_var_name_p i =
		if (i>= 0) && (i < !nb_variables) then get_var_name i
		else if (i>= !nb_variables) && (i< 2* !nb_variables) then
			(get_var_name (i- !nb_variables))^"'" 
		else failwith "get_var_name_p : incorrect index"
let print_const fmt (vect,cst,t) =
                let deb = ref true in
		for i=0 to (Array.length vect-1) do
		   if vect.(i) <> 0.0 then
		   begin
		     if vect.(i) > 0.0 then
		     begin
		       if (not !deb) then Format.pp_print_string fmt "+";
		       if (vect.(i)<> 1.0) then Format.fprintf fmt "%g*"
					 vect.(i)
		     end else begin
		       if (vect.(i)<> -1.0) then Format.fprintf fmt "%g*" 
								vect.(i)
		       else Format.pp_print_string fmt "-"
		     end;
		     Format.pp_print_string fmt (get_var_name_p i);
		     deb := false
		   end else if vect.(i) < 0.0 then
		   begin
		      Format.fprintf fmt "%g*%s" vect.(i)
				(get_var_name_p i);
		     deb := false
		   end
		done;
		if (cst <> 0.0) || (!deb) then
		begin
		    if (cst > 0.0) && (not !deb) then 
				Format.pp_print_string fmt "+";
		    Format.fprintf fmt "%g" cst
		end;
		Format.pp_print_string fmt (if t = EQ then "=0" else "<=0")


let print_etat fmt lst =
	List.iter 
		(fun a -> print_const fmt a; Format.pp_print_string fmt "\n")
			lst; Format.pp_print_string fmt ".\n"
let print_prog fmt (tinit,tsplit,lst) =
	Format.pp_print_string fmt "Init:\n"; 
	print_etat fmt tinit; 
	Format.pp_print_string fmt "Split:\n"; 
	print_etat fmt tsplit; 
	Format.pp_print_string fmt "\nTransitions:\n";
	List.iter (fun a -> print_etat fmt a) lst;
	Format.pp_print_string fmt "end."

	
